from ai2_kit.core.script import BashTemplate, BashStep, BashScript
from ai2_kit.core.artifact import Artifact, ArtifactDict
from ai2_kit.core.log import get_logger
from ai2_kit.core.job import gather_jobs
from ai2_kit.core.util import list_split, merge_dict

from typing import List, Optional
from pydantic import BaseModel
from dataclasses import dataclass

import numpy as np
import ase.io
import copy
import os
import re

from .iface import BaseCllContext, ICllExploreOutput
from .constant import DEFAULT_LASP_IN, DEFAULT_LAMMPS_TEMPLATE_FOR_DP_SSW, MODEL_DEVI_OUT
from .data import artifacts_to_ase_atoms, DataFormat


logger = get_logger(__name__)


class CllLaspInputConfig(BaseModel):
    class Potential(BaseModel):
        class LammpsPotential(BaseModel):
            input_template: str = DEFAULT_LAMMPS_TEMPLATE_FOR_DP_SSW

        lammps: Optional[LammpsPotential] = None  # currently only lammps is supported

    input_template: dict
    """
    Input template for LASP
    """
    potential: Potential

    system_files: List[str]
    """
    Initial system files to explore
    """


class CllLaspContextConfig(BaseModel):
    lasp_cmd: str = 'lasp'
    script_template: BashTemplate
    concurrency: int = 5


@dataclass
class CllLaspInput:
    config: CllLaspInputConfig
    type_map: List[str]
    mass_map: List[float]
    models: List[Artifact]
    new_system_files: List[Artifact]


@dataclass
class CllLaspContext(BaseCllContext):
    config: CllLaspContextConfig


@dataclass
class CllLaspOutput(ICllExploreOutput):
    output_dirs: List[Artifact]

    def get_model_devi_dataset(self) -> List[Artifact]:
        return self.output_dirs


async def cll_lasp(input: CllLaspInput, ctx: CllLaspContext):
    executor = ctx.resource_manager.default_executor

    # setup workspace
    work_dir = os.path.join(executor.work_dir, ctx.path_prefix)
    tasks_dir, = executor.setup_workspace(work_dir, ['tasks'])

    # resolve artifacts
    if len(input.new_system_files) > 0:
        systems = input.new_system_files
    else:
        systems = ctx.resource_manager.resolve_artifacts(input.config.system_files)

    # setup configuration
    lasp_in_data = merge_dict(copy.deepcopy(DEFAULT_LASP_IN), input.config.input_template)
    lasp_in_data['explore_type'] = 'ssw'
    lasp_in_data['SSW.output'] = 'T'

    lammps_input_template = None
    if input.config.potential.lammps is not None:
        lasp_in_data['potential'] = 'lammps'
        lammps_input_template = input.config.potential.lammps.input_template
    else:
        raise ValueError('At least one potential should be specified.')

    # make task dirs
    task_dirs = executor.run_python_fn(make_lasp_task_dirs)(
        systems=[a.to_dict() for a in systems],
        type_map=input.type_map, mass_map=input.mass_map,
        lasp_in_data=lasp_in_data,
        dp_models=[m.url for m in input.models],
        lammps_input_template=lammps_input_template,
        base_dir=tasks_dir,
    )

    # generate steps
    lasp_cmd = f'{ctx.config.lasp_cmd} 1 > lasp.out > lasp.err'
    steps = []
    for task_dir in task_dirs :
        steps.append(BashStep(cwd=task_dir['url'], cmd=lasp_cmd, checkpoint='lasp'))

    # submit jobs by the number of concurrency
    jobs = []
    for i, steps_group in enumerate(list_split(steps, ctx.config.concurrency)):
        if not steps_group:
            continue
        script = BashScript(
            template=ctx.config.script_template,
            steps=steps_group,
        )
        job = executor.submit(script.render(), cwd=tasks_dir,
                              checkpoint_key=f'queue-job/lasp/{i}:{tasks_dir}')
        jobs.append(job)
    await gather_jobs(jobs, max_tries=2)

    # process outputs
    executor.run_python_fn(process_lasp_outputs)(task_dirs=[a['url'] for a in task_dirs])

    output_dirs = [
        Artifact.of(
            url=task_dir['url'],
            executor=executor.name,
            format=DataFormat.LASP_LAMMPS_OUT_DIR,
            attrs={ **task_dir['attrs'],  'structures_file': 'structures.xyz'},
        ) for task_dir in task_dirs]  # type: ignore
    return CllLaspOutput(output_dirs=output_dirs)


def __export_remote_functions():

    from string import Template

    class LammpsInputTemplate(Template):
        delimiter = '$$'

    def make_lasp_task_dirs(systems: List[ArtifactDict],
                            lasp_in_data: dict,
                            base_dir: str,
                            type_map: List[str],
                            mass_map: List[float],
                            dp_models: List[str],
                            lammps_input_template: Optional[str],
                            ) -> List[ArtifactDict]:
        input_data = artifacts_to_ase_atoms(systems, type_map=type_map)

        i, task_dirs = 0, []  # TODO: why i is not generated from the loop?
        for artifact, atoms in input_data:
            # create task_dir
            task_dir = os.path.join(base_dir, f'task_{i:06}' )
            os.makedirs(task_dir, exist_ok=True)
            # create lasp.in
            lasp_in_text =  '\n'.join([f'{k:32} {v}' for k, v in lasp_in_data.items()])
            with open(os.path.join(task_dir, 'lasp.in'), 'w', encoding='utf-8') as f:
                f.write(lasp_in_text)
            # create lasp.str
            ase.io.write(os.path.join(task_dir, 'lasp.str'), atoms, format='dmol-arc')
            if lammps_input_template is not None:
                # create lammps.data
                data_file = os.path.join(task_dir, 'lammps.data')
                ase.io.write(data_file, atoms, format='lammps-data', specorder=type_map)  # type: ignore
                # create lammps input: in.simple, PS: Its LASP to blame for the name
                read_data_section = '\n'.join([
                    f"read_data {data_file}",
                    *(f"mass {i+1} {m}" for i, m in enumerate(mass_map))
                ])
                force_field_section = '\n'.join([
                    f"pair_style deepmd {' '.join(dp_models)} out_file {MODEL_DEVI_OUT}",
                    f"pair_coeff * *"
                ])
                lammps_input = LammpsInputTemplate(lammps_input_template).substitute(
                    read_data_section=read_data_section,
                    force_field_section=force_field_section,
                )
                lammps_input_file = os.path.join(task_dir, 'in.simple')
                with open(lammps_input_file, 'w', encoding='utf-8') as f:
                    f.write(lammps_input)
            else:
                raise ValueError('At least one potential should be specified.')
            # the `source` field is required as model_devi will use it to update init structures
            task_dirs.append({'url': task_dir,
                              'attrs': {**artifact['attrs'], 'source': artifact['url']}})

            i += 1  # TODO: refactor this
        return task_dirs


    def process_lasp_output(task_dir: str, file_name='structures.xyz'):
        """
        Align lasp output with model_devi records.

        As allstr.arc contains all the structures generated by LASP,
        we need to use the result of lasp.out for alignment.

        The following code is copy from ChecMate, may need to be refactored
        """
        all_str_file = os.path.join(task_dir, 'allstr.arc')
        lasp_out_file = os.path.join(task_dir, 'lasp.out')
        all_strs = ase.io.read(all_str_file, ':', format='dmol-arc')

        with open(all_str_file, "r") as f:
            all_qs = list((round(float(line[:73].strip().split()[-2]),6) for line in f.readlines() if "Energy" in line))
        with open(lasp_out_file, "r") as f:
            lines = f.readlines()
            traj_qs = list((round(float(line[:73].strip().split()[2]),6) for line in lines if "Energy,force" in line))
            traj_es = list((round(float(line[:73].strip().split()[1]),6) for line in lines if "Energy,force" in line))

        traj_strs = []
        for i, q in enumerate(all_qs):
            # FIXME: using fuzzy match to align data may have problem in some corner cases
            if len(traj_qs) > 0 and np.isclose(q, traj_qs[0], rtol=0, atol=0.0001):
                all_strs[i].info['ssw_energy'] = traj_es[len(traj_strs)]  # type: ignore
                traj_strs.append(all_strs[i])
                traj_qs.pop(0)
        # write trajectory to file
        ase.io.write(os.path.join(task_dir, file_name), traj_strs, format='extxyz')
        # edit model_devi.out
        model_devi_file = os.path.join(task_dir, MODEL_DEVI_OUT)
        lines = []
        with open(model_devi_file, "r") as f:
            for i, line in enumerate(f):
                if i > 0:
                    # replace step 0 with step i so that it can be aligned with structures
                    line = re.sub(r'^\s+\d+', f'{i-1:>12} ', line)   #
                lines.append(line)
        with open(model_devi_file, "w") as f:
            f.writelines(lines)

    def process_lasp_outputs(task_dirs: List[str], workers: int = 4):
        import joblib
        joblib.Parallel(n_jobs=workers)(
            joblib.delayed(process_lasp_output)(task_dir)
            for task_dir in task_dirs
        )


    return (
        make_lasp_task_dirs,
        process_lasp_outputs,
    )


(
    make_lasp_task_dirs,
    process_lasp_outputs,
) = __export_remote_functions()



from ai2_kit.core.artifact import Artifact, ArtifactDict
from ai2_kit.core.script import BashTemplate
from ai2_kit.core.script import BashScript, BashStep
from ai2_kit.core.job import JobFuture, gather_jobs
from ai2_kit.core.log import get_logger
from ai2_kit.core.util import dict_nested_get, expand_globs

from pydantic import BaseModel
from typing import List, Tuple
from dataclasses import dataclass
import os
import copy
import random
import sys
import json

from .iface import ICllTrainOutput, BaseCllContext
from .data import DataFormat, get_data_format

from .constant import (
    DP_CHECKPOINT_FILE,
    DP_DISP_FILE,
    DP_PROFILING_FILE,
    DP_INPUT_FILE,
    DP_FROZEN_MODEL,
    DP_ORIGINAL_MODEL
)

logger = get_logger(__name__)


class CllDeepmdInputConfig(BaseModel):

    model_num: int = 4
    """
    Total number of models to train.
    """
    init_dataset: List[str]
    """
    Dataset used to initialize training.
    """
    input_template: dict = dict()
    """
    Deepmd input template.
    """
    compress_model: bool = False
    """
    Whether to compress model after training.
    """
    isolate_outliers: bool = False
    """
    If isolate_outliers is enabled, then outlier data will be separated from training data.
    """
    outlier_f_cutoff: float = 10.
    """
    The threshold of force magnitude to determine whether a data is outlier.
    """
    outlier_weight: float = 0.003
    """
    The weight of outlier data in training data.
    """

    fixture_models: List[str] = []
    """
    Fixture models used to initialize training, support glob pattern.
    If this is not empty, then the whole training process will be skipped.
    This feature is useful for debugging, or explore more structures without training.
    The models should be on the remote executor.
    The name fixture is used as the concept of fixture in pytest.
    """

    group_by_formula: bool = False
    """
    Grouping dataset by formula
    If this is enabled, then the dataset will be grouped by formula.
    Otherwise, the dataset will be grouped by ancestor.

    Set this to True when you have multiple structures with the same ancestor.
    """


class CllDeepmdContextConfig(BaseModel):
    script_template: BashTemplate
    dp_cmd: str = 'dp'


@dataclass
class CllDeepmdInput:
    config: CllDeepmdInputConfig
    type_map: List[str]
    old_dataset: List[Artifact]  # training data used by previous iteration
    new_dataset: List[Artifact]  # training data used by current iteration


@dataclass
class CllDeepmdContext(BaseCllContext):
    config: CllDeepmdContextConfig


@dataclass
class GenericDeepmdOutput(ICllTrainOutput):
    models: List[Artifact]
    dataset: List[Artifact]

    def get_mlp_models(self) -> List[Artifact]:
        return self.models

    def get_training_dataset(self) -> List[Artifact]:
        return self.dataset


async def cll_deepmd(input: CllDeepmdInput, ctx: CllDeepmdContext):
    executor = ctx.resource_manager.default_executor

    # use fixture models
    if len(input.config.fixture_models) > 0:
        logger.info(f'Using fixture models: {input.config.fixture_models}')
        model_paths = executor.run_python_fn(expand_globs)(input.config.fixture_models)
        assert len(model_paths) > 0, f'No fixture models found: {input.config.fixture_models}'
        return GenericDeepmdOutput(
            dataset=input.old_dataset,
            models=[Artifact.of(url=url, format=DataFormat.DEEPMD_MODEL)
                    for url in model_paths]
        )

    # setup workspace
    work_dir = os.path.join(executor.work_dir, ctx.path_prefix)
    [new_dataset_dir, tasks_dir, outlier_dir] = executor.setup_workspace(
        work_dir, ['new_dataset', 'tasks', 'outlier_dataset'])

    init_dataset = ctx.resource_manager.resolve_artifacts(input.config.init_dataset)
    # input dataset contains data that generated by previous iteration
    # it should not contain the initial dataset
    input_dataset: List[Artifact] = input.old_dataset

    # make new dataset from raw data
    new_dataset, outlier_dataset = executor.run_python_fn(make_deepmd_dataset)(
        dataset_dir=new_dataset_dir,
        outlier_dir=outlier_dir,
        raw_data_collection=[a.to_dict() for a in input.new_dataset],
        type_map=input.type_map,
        isolate_outliers=input.config.isolate_outliers,
        outlier_f_cutoff=input.config.outlier_f_cutoff,
        group_by_formula=input.config.group_by_formula,
    )

    input_dataset += [ Artifact.of(**a) for a in new_dataset]
    # use attrs to distinguish outlier dataset
    input_dataset += [ Artifact.of(**{**a, 'attrs': {**a['attrs'], 'outlier': True}}) for a in outlier_dataset]

    #TODO: separate job creation and submission
    jobs: List[JobFuture] = []
    output_dirs = []

    # train multiple models at the same time
    for i in range(input.config.model_num):
        # each model should be trained in its own task_dir
        task_dir = os.path.join(tasks_dir, str(i).zfill(3))
        executor.mkdir(task_dir)

        # create dp train input file
        # NOTE: dp v1 format is supported currently
        # TODO: support more params if it is necessary

        # ref: https://github.com/deepmodeling/dpgen2/blob/master/examples/ch4/param_CH4_deepmd-kit-2.1.1.json
        # ref: https://github.com/deepmodeling/dpgen2/blob/master/dpgen2/op/prep_dp_train.py
        # ref: https://github.com/deepmodeling/dpgen2/blob/master/dpgen2/op/run_dp_train.py

        dp_input = copy.deepcopy(input.config.input_template)
        training: dict = dp_input['training']

        # set output files
        training['disp_file'] = DP_DISP_FILE
        training['save_ckpt'] = DP_CHECKPOINT_FILE
        training['profiling_file'] = DP_PROFILING_FILE

        # set random seed
        discriptor = dp_input['model']['descriptor']
        if discriptor['type'] == 'hybrid':
            for d in discriptor['list']:
                d['seed'] = _random_seed()
        else:
            discriptor['seed'] = _random_seed()
        dp_input['model']['fitting_net']['seed'] = _random_seed()
        dp_input['training']['seed'] = _random_seed()

        # set training data
        train_systems: List[str] = []
        outlier_systems: List[str] = []
        validation_systems: List[str] = []

        for a in input_dataset + init_dataset:
            if dict_nested_get(a.attrs, ['deepmd', 'validation_data'], False):
                validation_systems.append(a.url)
            else:
                if dict_nested_get(a.attrs, ['outlier'], False):
                    outlier_systems.append(a.url)
                else:
                    train_systems.append(a.url)


        auto_prob_str = "prob_sys_size"
        if input.config.isolate_outliers:
            # if isolate_outlier is enabled, then we need to set different weight for outlier data
            # e.g: prob_sys_size;0:57:0.9997;57:60:0.0003
            auto_prob_str += f';0:{len(train_systems)}:{1-input.config.outlier_weight}'
            auto_prob_str += f';{len(train_systems)}:{len(train_systems)+len(outlier_systems)}:{input.config.outlier_weight}'

        # Drop v1 support
        # TODO: remove in the future
        # training['systems'] = train_systems + outlier_systems
        # training.setdefault('set_prefix', 'set')  # respect user input
        # training.setdefault('batch_size', 'auto')
        # training['auto_prob_style'] = auto_prob_str

        # v2 training data
        training_data = training.get('training_data', {})
        training_data = {
            'set_prefix': 'set',
            'batch_size': 'auto',
            'sys_probs': None,
            ** training_data,
            'systems': train_systems + outlier_systems,
            'auto_prob_style': auto_prob_str,
        }
        training['training_data'] = training_data

        # ignore validation section if no data is provided, or else dp will throw error
        # OSError: [Errno cannot find valid a data system] Please check your setting for data systems
        if len(validation_systems) > 0:
            validation_data = {
                'systems': validation_systems,
                'set_prefix': training_data['set_prefix'],
                'batch_size': training_data['batch_size'],
            }
            training['validation_data'] = validation_data

        # other params
        dp_input['model']['type_map'] = input.type_map

        # write config to executor
        dp_input_text = json.dumps(dp_input, indent=2)
        dp_input_path = os.path.join(task_dir, DP_INPUT_FILE)
        executor.dump_text(dp_input_text, dp_input_path)

        # build script
        dp_cmd = ctx.config.dp_cmd
        dp_train_cmd = [dp_cmd, 'train', DP_INPUT_FILE]

        steps = [
            BashStep(cmd=dp_train_cmd, checkpoint='dp-train') # type: ignore
        ]

        if input.config.compress_model:
            dp_freeze_cmd = [dp_cmd, 'freeze', '-o', DP_ORIGINAL_MODEL]
            dp_compress_cmd = \
                [dp_cmd, 'compress', '-i', DP_ORIGINAL_MODEL, '-o', DP_FROZEN_MODEL]
            steps.append(BashStep(cmd=dp_freeze_cmd)) # type: ignore
            steps.append(BashStep(cmd=dp_compress_cmd)) # type: ignore
        else:
            dp_freeze_cmd = [dp_cmd, 'freeze', '-o', DP_FROZEN_MODEL]
            steps.append(BashStep(cmd=dp_freeze_cmd)) # type: ignore

        dp_train_script = BashScript(
            template=ctx.config.script_template,
            steps=steps # type: ignore
        )
        output_dirs.append(task_dir)

        # submit job
        job = executor.submit(dp_train_script.render(), cwd=task_dir, checkpoint_key=f'queue-job:dp-train:{task_dir}:{i}')
        jobs.append(job)

    await gather_jobs(jobs, max_tries=2)

    logger.info(f'All models are trained, output dirs: {output_dirs}')
    return GenericDeepmdOutput(
        dataset=input_dataset.copy(),
        models=[Artifact.of(
            url=os.path.join(url, DP_FROZEN_MODEL),
            format=DataFormat.DEEPMD_MODEL,
        ) for url in output_dirs]
    )


def _random_seed():
    return random.randrange(sys.maxsize) % (1 << 32)


def __export_remote_functions():
    import dpdata
    from itertools import groupby


    def make_deepmd_dataset(
        dataset_dir: str,
        outlier_dir: str,
        raw_data_collection: List[ArtifactDict],
        isolate_outliers: bool,
        outlier_f_cutoff: float,
        type_map: List[str],
        group_by_formula: bool = False,
    ):
        dataset_collection: List[Tuple[ArtifactDict, dpdata.LabeledSystem]] = []
        outlier_collection: List[Tuple[ArtifactDict, dpdata.LabeledSystem]] = []

        for raw_data in raw_data_collection:
            data_format = get_data_format(raw_data)  # type: ignore
            dp_system = None
            if data_format == DataFormat.CP2K_OUTPUT_DIR:
                dp_system = dpdata.LabeledSystem(os.path.join(raw_data['url'], 'output'), fmt='cp2k/output', type_map=type_map)
            elif data_format == DataFormat.VASP_OUTPUT_DIR:
                dp_system = dpdata.LabeledSystem(os.path.join(raw_data['url'], 'OUTCAR'), fmt='vasp/outcar', type_map=type_map)
            else:
                raise ValueError(f"Unsupported data format: {data_format}")
            # one case of len(dp_system) == 0 is when the system is not converged
            if dp_system is None or 0 == len(dp_system):
                continue  # skip invalid data

            if isolate_outliers and dp_system.data['forces'].max() > outlier_f_cutoff:
                outlier_collection.append((raw_data, dp_system))
            else:
                dataset_collection.append((raw_data, dp_system))

        _write_dp_dataset = _write_dp_dataset_by_formula if group_by_formula else _write_dp_dataset_by_ancestor

        dataset_dirs = _write_dp_dataset(dp_system_list=dataset_collection, out_dir=dataset_dir, type_map=type_map)
        outlier_dirs = _write_dp_dataset(dp_system_list=outlier_collection, out_dir=outlier_dir, type_map=type_map)

        return dataset_dirs, outlier_dirs


    def _write_dp_dataset_by_formula(dp_system_list: List[Tuple[ArtifactDict, dpdata.LabeledSystem]], out_dir: str, type_map: List[str]):
        """
        Write dp dataset that grouping by formula
        Use dpdata.MultipleSystems to merge systems with the same formula
        Use this when group by ancestor not works for you
        """
        if len(dp_system_list) == 0:
            return []
        os.makedirs(out_dir, exist_ok=True)

        multi_systems = dpdata.MultiSystems(dp_system_list[0][1])
        for _, system in dp_system_list[1:]:
            multi_systems.append(system)

        multi_systems.to_deepmd_npy(out_dir, type_map=type_map)  # type: ignore

        return [ {
            'url': os.path.join(out_dir, fname),
            'format': DataFormat.DEEPMD_NPY,
            'attrs': {},  # it is meaning less to set attrs in this case
        } for fname in os.listdir(out_dir)]


    def _write_dp_dataset_by_ancestor(dp_system_list: List[Tuple[ArtifactDict, dpdata.LabeledSystem]], out_dir: str, type_map: List[str]):
        """
        write dp dataset that grouping by ancestor
        """
        output_dirs: List[ArtifactDict] = []
        for key, dp_system_group in groupby(dp_system_list, key=lambda x: x[0]['attrs']['ancestor']):
            dp_system_group = list(dp_system_group)
            if 0 == len(dp_system_group):
                continue  # skip empty dataset
            group_out_dir = os.path.join(out_dir, key.replace('/', '_'))

            # merge dp_systems with the same ancestor into single data set
            dp_system = dp_system_group[0][1]
            for item in dp_system_group[1:]:
                dp_system += item[1]
            dp_system.to_deepmd_npy(group_out_dir, set_size=len(dp_system), type_map=type_map)  # type: ignore
            # inherit attrs key from input artifact
            output_dirs.append({'url': group_out_dir,
                                'format': DataFormat.DEEPMD_NPY,
                                'attrs': {**dp_system_group[0][0]['attrs']}})  # type: ignore
        return output_dirs


    return (make_deepmd_dataset, )

(
    make_deepmd_dataset,
) = __export_remote_functions()



from ai2_kit.core.util import ensure_dir
from ai2_kit.core.log import get_logger

import dpdata
import glob

logger = get_logger(__name__)

class DpdataHelper:

    def __init__(self, label: bool = True):
        self._systems = []
        self._label = label

    def read(self, *file_path_or_glob: str,**kwargs):
        kwargs.setdefault('fmt', 'deepmd/npy')
        files = []
        for file_path in file_path_or_glob:
            files += sorted(glob.glob(file_path, recursive=True)) if '*' in file_path else [file_path]

        if len(files) == 0:
            raise FileNotFoundError(f'No file found for {file_path_or_glob}')
        for file in files:
            self._read(file, **kwargs)
        return self

    def filter(self, lambda_expr: str):
        fn = eval(lambda_expr)
        def _fn(system):
            return fn(system.data)
        self._systems = list(filter(_fn, self._systems))
        return self

    @property
    def merge_write(self):
        logger.warn('merge_write is deprecated, use write instead')
        return self.write

    def write(self, out_path: str, fmt='deepmd/npy', merge: bool = True):
        ensure_dir(out_path)
        if len(self._systems) == 0:
            raise ValueError('No data to merge')

        if merge:
            systems = dpdata.MultiSystems(self._systems[0])
        else:
            if self._label:
                systems = dpdata.LabeledSystem(self._systems[0])
            else:
                systems = dpdata.System(self._systems[0])

        for system in self._systems[1:]:
            systems.append(system)

        if fmt == 'deepmd/npy':
            systems.to_deepmd_npy(out_path)  # type: ignore
        elif fmt == 'deepmd/raw':
            systems.to_deepmd_raw(out_path)  # type: ignore
        elif fmt == 'deepmd/hdf5':
            systems.to_deepmd_hdf5(out_path)  # type: ignore
        else:
            raise ValueError(f'Unknown fmt {fmt}')

    def _read(self, file: str, **kwargs):
        if self._label:
            self._systems.append(dpdata.LabeledSystem(file, **kwargs))
        else:
            self._systems.append(dpdata.System(file, **kwargs))



import numpy as np
import matplotlib.pyplot as plt
from typing import Optional


def display_lcurve(in_file: str = 'lcurve.out', save_to: Optional[str] = None, fig_ax = None):
    lcurve = np.loadtxt(in_file)
    step  = lcurve[:,0]  # step
    e_trn = lcurve[:,2]  # rmse_e_trn
    f_trn = lcurve[:,3]  # rmse_f_trn

    # plot e_trn and f_trn in log scale,
    # and make them fit the size of jupyter notebook
    if fig_ax is None:
        fig, ax = plt.subplots(1, 2, figsize=(12, 4), constrained_layout=True)
    else:
        fig, ax = fig_ax
    ax[0].plot(step, e_trn)
    ax[0].set_yscale('log')
    ax[0].set_xlabel(r'$STEP$')
    ax[0].set_ylabel(r'$RMSE_{E}$')
    ax[0].grid()
    ax[0].set_title('Learning Curve: RMSE of Energy')

    ax[1].plot(step, f_trn)
    ax[1].set_xlabel(r'$STEP$')
    ax[1].set_ylabel(r'$RMSE_{F}$')
    ax[1].grid()
    ax[1].set_title('Learning Curve: RMSE of Force')

    if save_to is None:
        fig.canvas.draw()
        fig.canvas.flush_events()
    else:
        fig.savefig(save_to, dpi=300, bbox_inches='tight')



from fabric import Connection, Result
from pydantic import BaseModel
from typing import Optional, List
from abc import ABC, abstractmethod
from io import StringIO
import numpy as np
import shlex
import invoke
import os
import stat
import json
import glob


class SshConfig(BaseModel):
    host: str
    gateway: Optional['SshConfig']


class BaseConnector(ABC):

    @abstractmethod
    def dump_text(self, text: str, path: str):
        ...

    @abstractmethod
    def glob(self, pattern: str) -> List[str]:
        ...

    @abstractmethod
    def run(self, script: str, **kwargs) -> Result:
        ...

    @abstractmethod
    def upload(self, from_path: str, to_dir: str) -> str:
        ...

    @abstractmethod
    def download(self, from_path: str, to_dir: str) -> str:
        ...

    @abstractmethod
    def sym_link(self, from_path: str, to_dir: str) -> str:
        ...

class SshConnector(BaseConnector):

    @classmethod
    def from_config(cls, config: SshConfig):
        connection = Connection(host=config.host)
        next_config = config.gateway
        next_connection = connection
        while next_config is not None:
            next_connection.gateway = Connection(host=next_config.host)
            next_connection = next_connection.gateway
            next_config = next_config.gateway
        return cls(connection)

    def __init__(self, connection: Connection):
        self._connection = connection

    def dump_text(self, text: str, path: str):
        f = StringIO(text)
        self.put(f, path)

    def glob(self, pattern: str):
        python_script = 'from glob import glob; from json import dumps; print(dumps(glob({})))'.format(repr(pattern))
        cmd = 'python -c {}'.format(shlex.quote(python_script))
        result = self.run(cmd, hide=True)
        return json.loads(result.stdout)

    def run(self, script, **kwargs) -> Result:
        return self._connection.run(script, **kwargs)

    def sym_link(self, from_path: str, to_dir: str) -> str:
        self.mkdir(to_dir)
        basename = safe_basename(from_path)
        to_path = os.path.join(to_dir, basename)
        self.run(get_ln_cmd(from_path, to_path))
        return to_path

    def upload(self, from_path: str, to_dir: str) -> str:
        self.mkdir(to_dir)
        to_path = os.path.join(to_dir, safe_basename(from_path))
        if os.path.isdir(from_path):
            self.put_dir(from_path, to_path)
        else:
            self.put(from_path, to_path)
        return to_path

    def download(self, from_path: str, to_dir: str) -> str:
        sftp = self.get_sftp()
        os.makedirs(to_dir, exist_ok=True)
        to_path = os.path.join(to_dir, safe_basename(from_path))
        if stat.S_ISDIR(sftp.lstat(from_path).st_mode):  # type: ignore
            self.get_dir(from_path, to_path)
        else:
            sftp.get(from_path, to_path)
        return to_path

    def put(self, *args, **kwargs):
        return self._connection.put(*args, **kwargs)

    def get(self, *args, **kwargs):
        return self._connection.get(*args, **kwargs)

    def mkdir(self, dir_path: str):
        self._connection.run('mkdir -p {}'.format(shlex.quote(dir_path)))

    def put_dir(self, from_dir: str, to_dir: str):
        self.mkdir(to_dir)
        for item in os.listdir(from_dir):
            from_path = os.path.join(from_dir, item)
            to_path = os.path.join(to_dir, item)
            if os.path.isdir(from_path):
                self.put_dir(from_path, to_path)
            else:
                self.put(from_path, to_path)

    def get_dir(self, from_dir: str, to_dir: str):
        sftp = self.get_sftp()
        os.makedirs(to_dir, exist_ok=True)

        for item in sftp.listdir_attr(from_dir):
            from_path = os.path.join(from_dir, item.filename)
            to_path = os.path.join(to_dir, item.filename)
            if stat.S_ISDIR(item.st_mode):  # type: ignore
                self.get_dir(from_path, to_path)
            else:
                sftp.get(from_path, to_path)

    def get_sftp(self):
        return self._connection.sftp()


class LocalConnector(BaseConnector):

    def dump_text(self, text: str, path: str):
        with open(path, 'w') as f:
            f.write(text)

    def glob(self, pattern: str):
        return glob.glob(pattern)

    def run(self, script, **kwargs):
        return invoke.run(script, **kwargs)

    def sym_link(self, from_path: str, to_dir: str) -> str:
        os.makedirs(to_dir, exist_ok=True)
        basename = safe_basename(from_path)
        to_path = os.path.join(to_dir, basename)
        self.run(get_ln_cmd(from_path, to_path))
        return to_path

    def upload(self, from_path: str, to_dir: str) -> str:
        return self.sym_link(from_path, to_dir)

    def download(self, from_path: str, to_dir: str) -> str:
        return self.sym_link(from_path, to_dir)


def get_ln_cmd(from_path: str, to_path: str):
    """
    The reason to `rm -d` to_path is to workaround the limit of ln.
    `ln` command cannot override existed directory,
    so we need to ensure to_path is not existed.
    Here we use -d option instead of -rf to avoid remove directory with content.
    The error of `rm -d` is suppressed as it will fail when to_path is file.
    `-T` option of `ln` is used to avoid some unexpected result.
    """
    to_path = os.path.normpath(to_path)
    return 'rm -d {to_path} || true && ln -sfT {from_path} {to_path}'.format(
        from_path=shlex.quote(from_path),
        to_path=shlex.quote(to_path)
    )

def safe_basename(path: str, default=''):
    """
    Ensure return valid file name as basename
    """
    basename = os.path.basename(path)
    if basename in ('/', '.', '..', ''):
        return default
    return basename



from pydantic import BaseModel
from typing import Optional, Dict
from abc import ABC, abstractmethod
from collections import defaultdict
import invoke
import shlex
import os
import re
import time
import asyncio


from .connector import BaseConnector
from .log import get_logger
from .job import JobFuture, JobState
from .checkpoint import apply_checkpoint, del_checkpoint
from .util import short_hash

logger = get_logger(__name__)


class QueueSystemConfig(BaseModel):
    class Slurm(BaseModel):
        sbatch_bin: str = 'sbatch'
        squeue_bin: str = 'squeue'
        scancel_bin: str = 'scancel'
        polling_interval: int = 10

    class LSF(BaseModel):
        bsub_bin: str = 'bsub'
        bjobs_bin: str = 'bjobs'
        polling_interval: int = 10

    slurm: Optional[Slurm]
    lsf: Optional[LSF]


class BaseQueueSystem(ABC):

    connector: BaseConnector

    @abstractmethod
    def get_polling_interval(self) -> int:
        ...

    @abstractmethod
    def get_script_suffix(self) -> str:
        ...

    @abstractmethod
    def get_submit_cmd(self) -> str:
        ...

    @abstractmethod
    def get_job_id_pattern(self) -> str:
        ...

    @abstractmethod
    def get_job_id_envvar(self) -> str:
        ...

    @abstractmethod
    def get_job_state(self, job_id: str, success_indicator_path: str) -> JobState:
        ...

    @abstractmethod
    def cancel(self, job_id: str):
        ...

    def _post_submit(self, job: 'QueueJobFuture'):
        ...

    def submit(self,
               script: str,
               cwd: str,
               name: Optional[str] = None,
               checkpoint_key: Optional[str] = None,
               success_indicator: Optional[str] = None,
               ):

        # use hash instead of uuid to ensure idempotence
        if name is None:
            name = 'job-' + short_hash(script) + self.get_script_suffix()
        quoted_cwd = shlex.quote(cwd)

        # a placeholder file that will be created when the script end without error
        if success_indicator is None:
            success_indicator = name + '.success'
        running_indicator = name + '.running'

        # TODO: maybe there are better way to inject running indicator
        script = inject_cmd_to_script(script, f'echo ${self.get_job_id_envvar()} > {shlex.quote(running_indicator)}')

        # create script and add a command to write job id to success indicator
        script = '\n'.join([
            script,
            '',
            f'echo ${self.get_job_id_envvar()} > {shlex.quote(success_indicator)}',
            '',
        ])

        script_path = os.path.join(cwd, name)
        self.connector.run(f'mkdir -p {quoted_cwd}')
        self.connector.dump_text(script, script_path)

        # submit script
        cmd = f"cd {quoted_cwd} && {self.get_submit_cmd()} {shlex.quote(name)}"

        # apply checkpoint
        submit_cmd_fn = self._submit_cmd
        if checkpoint_key is not None:
            submit_cmd_fn = apply_checkpoint(checkpoint_key)(submit_cmd_fn)

        # recover running job id
        # TODO: refactor the following code as function
        job_id, job_state  = None, JobState.UNKNOWN
        recover_cmd = f"cd {quoted_cwd} && cat {shlex.quote(running_indicator)}"
        try:
            job_id = self.connector.run(recover_cmd, hide=True).stdout.strip()
            if job_id:
                success_indicator_path = os.path.join(cwd, success_indicator)
                job_state = self.get_job_state(job_id, success_indicator_path=success_indicator_path)
        except:
            pass

        if job_id and job_state in (JobState.PENDING, JobState.RUNNING, JobState.COMPLETED):
            logger.info(f"{script_path} has been submmited ({job_id}) and in {str(job_state)} state, skip submit!")
        else:
            logger.info(f'Submit batch script: {script_path}')
            job_id = submit_cmd_fn(cmd)

        job = QueueJobFuture(self,
                             job_id=job_id,
                             name=name,
                             script=script,
                             cwd=cwd,
                             checkpoint_key=checkpoint_key,
                             success_indicator=success_indicator,
                             polling_interval=self.get_polling_interval() // 2,
                             )
        self._post_submit(job)
        return job

    def _submit_cmd(self, cmd: str):
        result = self.connector.run(cmd)
        m = re.search(self.get_job_id_pattern(), result.stdout)
        if m is None:
            raise RuntimeError("Unable to parse job id")
        return m.group(1)


class Slurm(BaseQueueSystem):
    config: QueueSystemConfig.Slurm

    _last_states: Optional[Dict[str, JobState]] = None
    _last_update_at: float = 0

    translate_table = {
        'PD': JobState.PENDING,
        'R': JobState.RUNNING,
        'CA': JobState.CANCELLED,
        'CF': JobState.PENDING,  # (configuring),
        'CG': JobState.RUNNING,  # (completing),
        'CD': JobState.COMPLETED,
        'F': JobState.FAILED,  # (failed),
        'TO': JobState.TIMEOUT,  # (timeout),
        'NF': JobState.FAILED,  # (node failure),
        'RV': JobState.FAILED,  # (revoked) and
        'SE': JobState.FAILED   # (special exit state)
    }

    def get_polling_interval(self):
        return self.config.polling_interval

    def get_script_suffix(self):
        return '.sbatch'

    def get_submit_cmd(self):
        return self.config.sbatch_bin

    def get_job_id_pattern(self):
        # example: Submitted batch job 123
        return r"Submitted batch job\s+(\d+)"

    def get_job_id_envvar(self) -> str:
        return 'SLURM_JOB_ID'

    def get_job_state(self, job_id: str, success_indicator_path: str) -> JobState:
        state = self._get_all_states().get(job_id)
        if state is None:
            cmd = 'test -f {}'.format(shlex.quote(success_indicator_path))
            ret = self.connector.run(cmd, warn=True)
            if ret.return_code:
                return JobState.FAILED
            else:
                return JobState.COMPLETED
        else:
            return state

    def cancel(self, job_id: str):
        cmd = '{} {}'.format(self.config.scancel_bin, job_id)
        self.connector.run(cmd)

    def _post_submit(self, job: 'QueueJobFuture'):
        self._last_states = None

    def _translate_state(self, slurm_state: str) -> JobState:
        return self.translate_table.get(slurm_state, JobState.UNKNOWN)

    def _get_all_states(self) -> Dict[str, JobState]:
        current_ts = time.time()
        # cache the states for 10 seconds to reduce the number of squeue calls
        if self._last_states is not None and (current_ts - self._last_update_at) < self.config.polling_interval:
            return self._last_states
        # call squeue to get all states
        cmd = f"{self.config.squeue_bin} --noheader --format='%i %t' -u $USER"
        try:
            r = self.connector.run(cmd, hide=True)
        except invoke.exceptions.UnexpectedExit as e:
            logger.warning(f'Error when calling squeue: {e}')
            return defaultdict(lambda: JobState.UNKNOWN) if self._last_states is None else self._last_states

        states: Dict[str, JobState] = dict()
        for line in r.stdout.splitlines():
            if not line:  # skip empty line
                continue
            job_id, slurm_state = line.split()
            state = self._translate_state(slurm_state)
            states[job_id] = state
        # update cache
        self._last_update_at = current_ts
        self._last_states = states
        return states


class Lsf(BaseQueueSystem):

    config: QueueSystemConfig.LSF

    def get_polling_interval(self):
        return self.config.polling_interval

    def get_script_suffix(self):
        return '.lsf'

    def get_submit_cmd(self):
        return self.config.bsub_bin + ' <'

    def get_job_id_pattern(self):
        # example: Job <123> is submitted to queue <small>.
        return r"Job <(\d+)> is submitted to queue"

    def get_job_id_envvar(self) -> str:
        return 'LSB_JOBID'

    # TODO
    def get_job_state(self, job_id: str, success_indicator_path: str) -> JobState:
        return JobState.UNKNOWN

    # TODO
    def cancel(self, job_id: str):
        ...


class QueueJobFuture(JobFuture):

    def __init__(self,
                 queue_system: BaseQueueSystem,
                 job_id: str,
                 script: str,
                 cwd: str,
                 name: str,
                 success_indicator: str,
                 checkpoint_key: Optional[str],
                 polling_interval=10,
                 ):
        self._queue_system = queue_system
        self._name = name
        self._script = script
        self._cwd = cwd
        self._job_id = job_id
        self._checkpoint_key = checkpoint_key
        self._success_indicator = success_indicator
        self._polling_interval = polling_interval
        self._final_state = None

    @property
    def success_indicator_path(self):
        return os.path.join(self._cwd, self._success_indicator)

    def get_job_state(self):
        if self._final_state is not None:
            return self._final_state

        state = self._queue_system.get_job_state(
            self._job_id, self.success_indicator_path)
        if state.terminal:
            self._final_state = state

        return state

    def resubmit(self):
        if not self.done():
            raise RuntimeError('Cannot resubmit an unfinished job!')

        # delete checkpoint on resubmit
        if self._checkpoint_key is not None:
            logger.info(f'Delete checkpoint due to resubmit {self._checkpoint_key}')
            del_checkpoint(self._checkpoint_key)

        logger.info(f'Resubmit job: {self._job_id}')
        return self._queue_system.submit(
            script=self._script,
            cwd=self._cwd,
            name=self._name,
            checkpoint_key=self._checkpoint_key,
            success_indicator=self._success_indicator,
        )

    def is_success(self):
        return self.get_job_state() is JobState.COMPLETED

    def cancel(self):
        self._queue_system.cancel(self._job_id)

    def done(self):
        return self.get_job_state().terminal

    def result(self, timeout: float = float('inf')) -> JobState:
        return asyncio.run(self.result_async(timeout))

    async def result_async(self, timeout: float = float('inf')) -> JobState:
        '''
        Though this is not fully async, as the job submission and state polling are still blocking,
        but it is already good enough to handle thousands of jobs (I guess).
        '''
        timeout_ts = time.time() + timeout
        while time.time() < timeout_ts:
            if self.done():
                return self.get_job_state()
            else:
                await asyncio.sleep(self._polling_interval)
        else:
            raise RuntimeError(f'Timeout of polling job: {self._job_id}')

    def __repr__(self):
        return repr(dict(
            name=self._name,
            cwd=self._cwd,
            job_id=self._job_id,
            success_indicator=self._success_indicator,
            polling_interval=self._polling_interval,
            state=self.get_job_state(),
        ))


def inject_cmd_to_script(script: str, cmd: str):
    """
    Find the position of first none comment or empty lines,
    and inject command before it
    """
    lines = script.splitlines()
    i = 0
    for i, line in enumerate(lines):
        line = line.strip()
        if line and not line.startswith('#'):
            break
    lines.insert(i, cmd)
    return '\n'.join(lines)
